/*  fwdmodel_dwi.cc - Implements a convolution based model for DWI analysis

    Jesper Kallehauge, IBME

    Copyright (C) 2016 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#include "fwdmodel_dwi.h"

#include "newimage/newimageall.h"
#include <iostream>
#include "armawrap/newmat.h"
#include <stdexcept>

#include "fabber_core/easylog.h"
#include "miscmaths/miscprob.h"

using namespace std;
using namespace NEWMAT;
using namespace NEWIMAGE;


FactoryRegistration<FwdModelFactory, DWIFwdModel>
    DWIFwdModel::registration("dwi");

static OptionSpec OPTIONS[] = {
    { "bvals", OPT_MATRIX, "ASCII matrix containing bvals", OPT_REQ, "" },
    { "" },
};

void DWIFwdModel::GetOptions(vector<OptionSpec> &opts) const
{
    for (int i = 0; OPTIONS[i].name != ""; i++)
    {
        opts.push_back(OPTIONS[i]);
    }
}

std::string DWIFwdModel::GetDescription() const
{
    return string("One-compartment model with two parameters: ")
        + "ADC the Apparent Diffusion Coefficient and "
        + "sig0 the initial signal";
}

string DWIFwdModel::ModelVersion() const
{
    string version = "fwdmodel_dwi.cc";
#ifdef GIT_SHA1
    version += string(" Revision ") + GIT_SHA1;
#endif
#ifdef GIT_DATE
    version += string(" Last commit ") + GIT_DATE;
#endif
    return version;
}

void DWIFwdModel::HardcodedInitialDists(MVNDist &prior,
    MVNDist &posterior) const
{
    assert(prior.means.Nrows() == NumParams());

    SymmetricMatrix precisions = IdentityMatrix(NumParams()) * 1e-12;

    // Set priors
    // Fp or Ktrans whatever you belive
    prior.means(ADC_index()) = 0.01;
    precisions(ADC_index(), ADC_index()) = 1e-12;

    prior.means(sig0_index()) = 0.01;
    precisions(sig0_index(), sig0_index()) = 1e-12;

    // Set precsions on priors
    prior.SetPrecisions(precisions);

    // Set initial posterior
    posterior = prior;

    // For parameters with uniformative prior chosoe more sensible inital posterior
    // Tissue perfusion
    posterior.means(ADC_index()) = 1;
    precisions(ADC_index(), ADC_index()) = 0.1;

    posterior.SetPrecisions(precisions);
}

void DWIFwdModel::Evaluate(const ColumnVector &params, ColumnVector &result) const
{
    // ensure that values are reasonable
    // negative check
    ColumnVector paramcpy = params;
    for (int i = 1; i <= NumParams(); i++)
    {
        if (params(i) < 0)
        {
            paramcpy(i) = 0;
        }
    }

    // parameters that are inferred - extract and give sensible names
    float ADC;
    float sig0; //'inital' value of the signal

    // extract values from params
    sig0 = paramcpy(sig0_index());
    ADC = paramcpy(ADC_index());
    if (ADC < 1e-8)
        ADC = 1e-8;
    if (sig0 < 1e-8)
        sig0 = 1e-8;

    ColumnVector bvalshere; // the arterial signal to use for the analysis
    if (bvals.Nrows() > 0)
    {
        bvalshere = bvals; //use the artsig that was loaded by the model
    }
    else
    {
        //use an artsig from supplementary data
        if (suppdata.Nrows() > 0)
        {
            bvalshere = suppdata;
        }
        else
        {
            cout << "No valid b values found" << endl;
            throw;
        }
    }
    // use length of the aif to determine the number of time points
    int ntpts = bvalshere.Nrows();

    // --- Residue Function ----
    ColumnVector sig(ntpts);
    sig = 0.0;

    sig = sig0 * MISCMATHS::exp(-ADC * 1e-3 * bvals); // the 1 compartment model.
    //cout<<sig<<endl;
    result = sig;

    for (int i = 1; i <= ntpts; i++)
    {
        if (isnan(result(i)) || isinf(result(i)))
        {
            exit(0);
            LOG << "Warning NaN of inf in result" << endl;
            LOG << "result: " << result.t() << endl;
            LOG << "params: " << params.t() << endl;

            result = 0.0;
            break;
        }
    }
}

FwdModel *DWIFwdModel::NewInstance()
{
    return new DWIFwdModel();
}

void DWIFwdModel::Initialize(ArgsType &args)
{
    string bvalfile = args.Read("bvals");
    if (bvalfile != "none")
    {
        bvals = MISCMATHS::read_ascii_matrix(bvalfile);
    }
    doard = false;
}

vector<string> DWIFwdModel::GetUsage() const
{
    vector<string> usage;

    usage.push_back("\nThis model is a one compartment model\n");
    usage.push_back("It returns  2 parameters :\n");
    usage.push_back(" ADC: the Apparent Diffusion Coefficient\n");
    usage.push_back(" sig0: the initial signal\n");

    return usage;
}

void DWIFwdModel::NameParams(vector<string> &names) const
{
    names.clear();

    names.push_back("ADC");
    names.push_back("sig0");
}
